from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Measurement, Total, Units
from solax.utils import div10, div100, pack_u16, to_signed, twoway_div10, twoway_div100


class X3V34(Inverter):
    """X3 v2.034.06"""

    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(int, 5),
            vol.Required("sn"): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=200, max=200),
                )
            ),
            vol.Required("information"): vol.Schema(
                vol.All(vol.Length(min=10, max=10))
            ),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def response_decoder(cls):
        return {
            "Network Voltage Phase 1": (0, Units.V, div10),
            "Network Voltage Phase 2": (1, Units.V, div10),
            "Network Voltage Phase 3": (2, Units.V, div10),
            "Output Current Phase 1": (3, Units.A, twoway_div10),
            "Output Current Phase 2": (4, Units.A, twoway_div10),
            "Output Current Phase 3": (5, Units.A, twoway_div10),
            "Power Now Phase 1": (6, Units.W, to_signed),
            "Power Now Phase 2": (7, Units.W, to_signed),
            "Power Now Phase 3": (8, Units.W, to_signed),
            "PV1 Voltage": (9, Units.V, div10),
            "PV2 Voltage": (10, Units.V, div10),
            "PV1 Current": (11, Units.A, div10),
            "PV2 Current": (12, Units.A, div10),
            "PV1 Power": (13, Units.W),
            "PV2 Power": (14, Units.W),
            "Total PV Energy": (pack_u16(89, 90), Total(Units.KWH), div10),
            "Today's PV Energy": (112, DailyTotal(Units.KWH), div10),
            "Grid Frequency Phase 1": (15, Units.HZ, div100),
            "Grid Frequency Phase 2": (16, Units.HZ, div100),
            "Grid Frequency Phase 3": (17, Units.HZ, div100),
            "Total Energy": (pack_u16(19, 20), Total(Units.KWH), div10),
            "Today's Energy": (21, DailyTotal(Units.KWH), div10),
            "Battery Voltage": (24, Units.V, div100),
            "Battery Current": (25, Units.A, twoway_div100),
            "Battery Power": (26, Units.W, to_signed),
            "Battery Temperature": (27, Units.C),
            "Battery Remaining Capacity": (28, Units.PERCENT),
            "Total Battery Discharge Energy": (
                pack_u16(30, 31),
                Total(Units.KWH),
                div10,
            ),
            "Today's Battery Discharge Energy": (113, DailyTotal(Units.KWH), div10),
            "Battery Remaining Energy": (
                32,
                Measurement(Units.KWH, storage=True),
                div10,
            ),
            "Total Battery Charge Energy": (
                pack_u16(87, 88),
                Total(Units.KWH),
                div10,
            ),
            "Today's Battery Charge Energy": (114, DailyTotal(Units.KWH), div10),
            "Exported Power": (65, Units.W, to_signed),
            "Total Feed-in Energy": (pack_u16(67, 68), Total(Units.KWH), div100),
            "Total Consumption": (pack_u16(69, 70), Total(Units.KWH), div100),
            "AC Power": (181, Units.W, to_signed),
            "EPS Frequency": (63, Units.HZ, div100),
            "EPS Total Energy": (
                pack_u16(110, 111),
                Measurement(Units.KWH, storage=False),
                div10,
            ),
        }

    # pylint: enable=duplicate-code

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]



from collections import namedtuple
from copy import copy

import pytest

import solax.inverters as inverter
from tests.samples.expected_values import (
    QVOLTHYBG33P_VALUES,
    X1_BOOST_VALUES,
    X1_BOOST_VALUES_OVERFLOWN,
    X1_HYBRID_G4_VALUES,
    X1_MINI_VALUES,
    X1_MINI_VALUES_V34,
    X1_MINI_VALUES_V34_VER3,
    X1_SMART_VALUES,
    X1_VALUES,
    X3_HYBRID_G4_VALUES,
    X3_HYBRID_VALUES,
    X3_MICPRO_G2_VALUES,
    X3_ULTRA_VALUES,
    X3_VALUES,
    X3V34_HYBRID_VALUES,
    X3V34_HYBRID_VALUES_EPS_MODE,
    X3V34_HYBRID_VALUES_NEGATIVE_POWER,
    XHYBRID_VALUES,
)
from tests.samples.responses import (
    QVOLTHYBG33P_RESPONSE_V34,
    X1_BOOST_AIR_MINI_RESPONSE,
    X1_BOOST_RESPONSE,
    X1_BOOST_RESPONSE_OVERFLOWN,
    X1_HYBRID_G3_2X_MPPT_RESPONSE,
    X1_HYBRID_G3_RESPONSE,
    X1_HYBRID_G4_RESPONSE,
    X1_MINI_RESPONSE_V34,
    X1_MINI_RESPONSE_V34_VER3,
    X1_SMART_RESPONSE,
    X3_HYBRID_G3_2X_MPPT_RESPONSE,
    X3_HYBRID_G3_2X_MPPT_RESPONSE_V34,
    X3_HYBRID_G3_2X_MPPT_RESPONSE_V34_EPS_MODE,
    X3_HYBRID_G3_2X_MPPT_RESPONSE_V34_NEGATIVE_POWER,
    X3_HYBRID_G3_RESPONSE,
    X3_HYBRID_G4_RESPONSE,
    X3_MIC_RESPONSE,
    X3_MICPRO_G2_RESPONSE,
    X3_ULTRA_RESPONSE,
    XHYBRID_DE01_RESPONSE,
    XHYBRID_DE02_RESPONSE,
)

X_FORWARDED_HEADER = {"X-Forwarded-For": "5.8.8.8"}


@pytest.fixture()
def simple_http_fixture(httpserver):
    httpserver.expect_request(
        uri="/", method="GET", query_string="index.html"
    ).respond_with_json({"hello": "world"})
    yield (httpserver.host, httpserver.port)


InverterUnderTest = namedtuple(
    "InverterUnderTest",
    "uri, method, query_string, response, inverter, values, headers, data",
)

INVERTERS_UNDER_TEST = [
    InverterUnderTest(
        uri="/api/realTimeData.htm",
        method="GET",
        query_string=None,
        response=XHYBRID_DE01_RESPONSE,
        inverter=inverter.XHybrid,
        values=XHYBRID_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/api/realTimeData.htm",
        method="GET",
        query_string=None,
        response=XHYBRID_DE02_RESPONSE,
        inverter=inverter.XHybrid,
        values=XHYBRID_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string=None,
        response=X1_HYBRID_G4_RESPONSE,
        inverter=inverter.X1HybridGen4,
        values=X1_HYBRID_G4_VALUES,
        headers=None,
        data="optType=ReadRealTimeData",
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_BOOST_AIR_MINI_RESPONSE,
        inverter=inverter.X1Mini,
        values=X1_MINI_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_MINI_RESPONSE_V34,
        inverter=inverter.X1MiniV34,
        values=X1_MINI_VALUES_V34,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_SMART_RESPONSE,
        inverter=inverter.X1Smart,
        values=X1_SMART_VALUES,
        headers=X_FORWARDED_HEADER,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_BOOST_RESPONSE,
        inverter=inverter.X1Boost,
        values=X1_BOOST_VALUES,
        headers=X_FORWARDED_HEADER,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_BOOST_RESPONSE_OVERFLOWN,
        inverter=inverter.X1Boost,
        values=X1_BOOST_VALUES_OVERFLOWN,
        headers=X_FORWARDED_HEADER,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_MIC_RESPONSE,
        inverter=inverter.X3,
        values=X3_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_HYBRID_G3_RESPONSE,
        inverter=inverter.X3,
        values=X3_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_HYBRID_G3_RESPONSE,
        inverter=inverter.X1,
        values=X1_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X1_HYBRID_G3_2X_MPPT_RESPONSE,
        inverter=inverter.X1,
        values=X1_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_HYBRID_G3_2X_MPPT_RESPONSE,
        inverter=inverter.X3,
        values=X3_HYBRID_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_HYBRID_G3_2X_MPPT_RESPONSE_V34,
        inverter=inverter.X3V34,
        values=X3V34_HYBRID_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_HYBRID_G3_2X_MPPT_RESPONSE_V34_NEGATIVE_POWER,
        inverter=inverter.X3V34,
        values=X3V34_HYBRID_VALUES_NEGATIVE_POWER,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="optType=ReadRealTimeData",
        response=X3_HYBRID_G3_2X_MPPT_RESPONSE_V34_EPS_MODE,
        inverter=inverter.X3V34,
        values=X3V34_HYBRID_VALUES_EPS_MODE,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string=None,
        response=X3_HYBRID_G4_RESPONSE,
        inverter=inverter.X3HybridG4,
        values=X3_HYBRID_G4_VALUES,
        headers=None,
        data="optType=ReadRealTimeData",
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string=None,
        response=X3_MICPRO_G2_RESPONSE,
        inverter=inverter.X3MicProG2,
        values=X3_MICPRO_G2_VALUES,
        headers=None,
        data="optType=ReadRealTimeData",
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string=None,
        response=X3_ULTRA_RESPONSE,
        inverter=inverter.X3Ultra,
        values=X3_ULTRA_VALUES,
        headers=None,
        data="optType=ReadRealTimeData",
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string="",
        response=QVOLTHYBG33P_RESPONSE_V34,
        inverter=inverter.QVOLTHYBG33P,
        values=QVOLTHYBG33P_VALUES,
        headers=None,
        data=None,
    ),
    InverterUnderTest(
        uri="/",
        method="POST",
        query_string=None,
        response=X1_MINI_RESPONSE_V34_VER3,
        inverter=inverter.X1MiniV34,
        values=X1_MINI_VALUES_V34_VER3,
        headers=None,
        data="optType=ReadRealTimeData",
    ),
]


@pytest.fixture(params=INVERTERS_UNDER_TEST)
def inverters_under_test(request):
    yield request.param.inverter


@pytest.fixture(params=INVERTERS_UNDER_TEST)
def inverters_fixture(httpserver, request):
    httpserver.expect_request(
        uri=request.param.uri,
        method=request.param.method,
        query_string=request.param.query_string,
        headers=request.param.headers,
        data=request.param.data,
    ).respond_with_json(request.param.response)
    yield (
        (httpserver.host, httpserver.port),
        request.param.inverter,
        request.param.values,
    )


@pytest.fixture(params=INVERTERS_UNDER_TEST)
def inverters_garbage_fixture(httpserver, request):
    httpserver.expect_request(
        uri=request.param.uri,
        method=request.param.method,
        query_string=request.param.query_string,
    ).respond_with_json({"bingo": "bango"})
    yield ((httpserver.host, httpserver.port), request.param.inverter)


@pytest.fixture(params=INVERTERS_UNDER_TEST)
def inverters_fixture_all_zero(httpserver, request):
    """Use defined responses but replace the data with all zero values.
    Testing incorrect responses.
    """

    response = request.param.response
    response = copy(response)
    response["Data"] = [0] * (len(response["Data"]))

    httpserver.expect_request(
        uri=request.param.uri,
        method=request.param.method,
        query_string=request.param.query_string,
        headers=request.param.headers,
        data=request.param.data,
    ).respond_with_json(response)
    yield (
        (httpserver.host, httpserver.port),
        request.param.inverter,
        request.param.values,
    )



from __future__ import annotations

import dataclasses
import sys
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional
from weakref import WeakValueDictionary

import aiohttp

__all__ = ("InverterHttpClient", "Method")

if sys.version_info >= (3, 10):
    from dataclasses import KW_ONLY


REQUEST_TIMEOUT = 5.0
_CACHE: WeakValueDictionary[int, InverterHttpClient] = WeakValueDictionary()


class Method(Enum):
    GET = 1
    POST = 2


_kwargs: Dict[str, bool] = {}

if sys.version_info >= (3, 11):
    _kwargs["slots"] = True
    _kwargs["weakref_slot"] = True


@dataclass(frozen=True, **_kwargs)
class InverterHttpClient:
    """Initialize the Http client."""

    if sys.version_info >= (3, 10):
        _: KW_ONLY

    url: str
    method: Method
    pwd: str
    headers: Dict[str, str] = field(default_factory=dict)
    data: Optional[bytes] = None
    query: str = ""

    def __hash__(self):
        return id(self)

    def replace(self, **kwargs) -> InverterHttpClient:
        fields = dataclasses.fields(InverterHttpClient)
        data = {}
        values = []

        for fld in fields:
            if fld.name in kwargs:
                value = kwargs.pop(fld.name)
            else:
                value = getattr(self, fld.name)

            data[fld.name] = value

            if isinstance(value, dict):
                value = dict(value)
                values.append(tuple(value.items()))
            else:
                values.append(value)

            data[fld.name] = value

        key = hash(tuple(values))
        cached = _CACHE.get(key)

        if cached is None:
            cached = _CACHE[key] = InverterHttpClient(**data)

        return cached

    def with_headers(self, headers) -> InverterHttpClient:
        return self.replace(headers=dict(headers))

    def with_default_data(self) -> InverterHttpClient:
        data = "optType=ReadRealTimeData"
        if self.pwd:
            data = data + "&pwd=" + self.pwd
        return self.with_data(data)

    def with_data(self, data) -> InverterHttpClient:
        return self.replace(data=data)

    def with_query(self, query) -> InverterHttpClient:
        return self.replace(query=query)

    def with_default_query(self) -> InverterHttpClient:
        if self.pwd:
            base = "optType=ReadRealTimeData&pwd={}&"
            query = base.format(self.pwd)
        else:
            query = "optType=ReadRealTimeData"

        return self.with_query(query)

    async def request(self):
        if self.method is Method.POST:
            return await self.post()
        return await self.get()

    async def get(self):
        url = self.url + "?" + self.query if self.query else self.url
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, headers=self.headers, timeout=REQUEST_TIMEOUT
            ) as req:
                req.raise_for_status()
                resp = await req.read()
        return resp

    async def post(self):
        url = self.url + "?" + self.query if self.query else self.url
        data = self.data.encode("utf-8") if self.data else None
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, headers=self.headers, data=data, timeout=REQUEST_TIMEOUT
            ) as req:
                req.raise_for_status()
                resp = await req.read()
        return resp

    def __str__(self) -> str:
        using = "query in url" if self.query else "data in the body"
        return f"{self.url} using {using}"



# pylint: disable=unused-import
from tests.fixtures import inverters_fixture  # noqa: F401
from tests.fixtures import inverters_fixture_all_zero  # noqa: F401
from tests.fixtures import inverters_garbage_fixture  # noqa: F401
from tests.fixtures import inverters_under_test  # noqa: F401
from tests.fixtures import simple_http_fixture  # noqa: F401



from abc import abstractmethod
from typing import Any, Dict, Optional, Tuple

import aiohttp
import voluptuous as vol

from solax import utils
from solax.inverter_http_client import InverterHttpClient, Method
from solax.response_parser import InverterResponse, ResponseDecoder, ResponseParser
from solax.units import Measurement, Units


class InverterError(Exception):
    """Indicates error communicating with inverter"""


class Inverter:
    """Base wrapper around Inverter HTTP API"""

    @classmethod
    def response_decoder(cls) -> ResponseDecoder:
        """
        Inverter implementations should override
        this to return a decoding map
        """
        raise NotImplementedError()

    # pylint: enable=C0301
    _schema = vol.Schema({})  # type: vol.Schema

    def __init__(self, http_client: InverterHttpClient):
        self.manufacturer = "Solax"
        self.http_client = http_client

        schema = type(self).schema()
        response_decoder = type(self).response_decoder()
        dongle_serial_number_getter = type(self).dongle_serial_number_getter
        inverter_serial_number_getter = type(self).inverter_serial_number_getter
        self.response_parser = ResponseParser(
            schema,
            response_decoder,
            dongle_serial_number_getter,
            inverter_serial_number_getter,
        )

    @classmethod
    def _build(cls, host, port, pwd="", params_in_query=True):
        url = utils.to_url(host, port)
        http_client = InverterHttpClient(url=url, method=Method.POST, pwd=pwd)
        if params_in_query:
            http_client = http_client.with_default_query()
        else:
            http_client = http_client.with_default_data()

        return cls(http_client)

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        versions = {
            cls._build(host, port, pwd, True),
            cls._build(host, port, pwd, False),
        }
        return versions

    async def get_data(self) -> InverterResponse:
        try:
            data = await self.make_request()
        except aiohttp.ClientError as ex:
            msg = "Could not connect to inverter endpoint"
            raise InverterError(msg, str(self.__class__.__name__)) from ex
        except vol.Invalid as ex:
            msg = "Received malformed JSON from inverter"
            raise InverterError(msg, str(self.__class__.__name__)) from ex
        return data

    async def make_request(self) -> InverterResponse:
        """
        Return instance of 'InverterResponse'
        Raise exception if unable to get data
        """
        raw_response = await self.http_client.request()
        return self.response_parser.handle_response(raw_response)

    @classmethod
    def sensor_map(cls) -> Dict[str, Tuple[int, Measurement]]:
        """
        Return sensor map
        Warning, HA depends on this
        """
        sensors: Dict[str, Tuple[int, Measurement]] = {}
        for name, mapping in cls.response_decoder().items():
            unit = Measurement(Units.NONE)

            (idx, unit_or_measurement, *_) = mapping

            if isinstance(unit_or_measurement, Units):
                unit = Measurement(unit_or_measurement)
            else:
                unit = unit_or_measurement
            if isinstance(idx, tuple):
                sensor_indexes = idx[0]
                first_sensor_index = sensor_indexes[0]
                idx = first_sensor_index
            sensors[name] = (idx, unit)
        return sensors

    @classmethod
    def schema(cls) -> vol.Schema:
        """
        Return schema
        """
        return cls._schema

    @classmethod
    def dongle_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["sn"]

    @classmethod
    @abstractmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        raise NotImplementedError  # pragma: no cover

    def __str__(self) -> str:
        return f"{self.__class__.__name__}::{self.http_client}"



import pytest

import solax
from solax.discovery import REGISTRY
from solax.inverter import Inverter, InverterError
from solax.units import Measurement
from tests import fixtures


async def build_right_variant(inverter, conn) -> Inverter:
    last_error: BaseException = BaseException("anticipating errors")
    for i in inverter.build_all_variants(*conn):
        try:
            await i.get_data()
            return i
        except InverterError as ex:
            last_error = ex
    raise last_error


@pytest.mark.asyncio
async def test_smoke(inverters_fixture):
    conn, inverter_class, values = inverters_fixture
    inv = await build_right_variant(inverter_class, conn)
    rt_api = solax.RealTimeAPI(inv)
    parsed = await rt_api.get_data()

    msg = "data size should match expected values"
    assert len(values) == len(parsed.data), msg
    for sensor, value in values.items():
        assert (
            parsed.data[sensor] == value
        ), f"{sensor}: expected {value} but got {parsed.data[sensor]}"


@pytest.mark.asyncio
async def test_throws_when_unable_to_parse(inverters_garbage_fixture):
    conn, inverter_class = inverters_garbage_fixture
    with pytest.raises(InverterError):
        i = await build_right_variant(inverter_class, conn)
        await i.get_data()


def test_registry_matches_inverters_under_test():
    test_inverters = {i.inverter for i in fixtures.INVERTERS_UNDER_TEST}
    registry_inverters = set(REGISTRY)
    assert test_inverters == registry_inverters, "tests do not match registry"


def test_inverter_sensors_match():
    test_values = ((i.inverter, i.values) for i in fixtures.INVERTERS_UNDER_TEST)
    for i, expected_values in test_values:
        sensor_map = i.sensor_map()
        msg = f"""{sorted(sensor_map.keys())} vs
{sorted(expected_values.keys())}"""
        assert len(sensor_map) == len(expected_values), msg
        for name, _ in sensor_map.items():
            assert name in expected_values


def test_inverter_sensors_define_valid_units(inverters_under_test):
    sensor_map = inverters_under_test.sensor_map()
    for name, (_, unit, *_) in sensor_map.items():
        msg = (
            f"provided unit '{unit}'({type(unit)}) "
            f"is not a proper Unit on sensor '{name}' of Inverter '{inverters_under_test}'"
        )
        assert isinstance(unit, Measurement), msg


@pytest.mark.asyncio
async def test_smoke_zero(inverters_fixture_all_zero):
    """Responses with all zero values should be treated as an error.
    Args:
        inverters_fixture_all_zero (_type_): all responses with zero value data
    """
    conn, inverter_class, _ = inverters_fixture_all_zero

    # msg = 'all zero values should be discarded'
    with pytest.raises(InverterError):
        inv = await build_right_variant(inverter_class, conn)
        rt_api = solax.RealTimeAPI(inv)
        await rt_api.get_data()



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Total, Units
from solax.utils import div10, div100, pack_u16, to_signed


class X1HybridGen4(Inverter):
    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(int, 15),
            vol.Required(
                "sn",
            ): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=200, max=200),
                )
            ),
            vol.Required("information"): vol.Schema(vol.All(vol.Length(min=9, max=10))),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        versions = [cls._build(host, port, pwd, False)]
        return versions

    @classmethod
    def response_decoder(cls):
        return {
            "AC voltage R": (0, Units.V, div10),
            "AC current": (1, Units.A, div10),
            "AC power": (2, Units.W, to_signed),
            "Grid frequency": (3, Units.HZ, div100),
            "PV1 voltage": (4, Units.V, div10),
            "PV2 voltage": (5, Units.V, div10),
            "PV1 current": (6, Units.A, div10),
            "PV2 current": (7, Units.A, div10),
            "PV1 power": (8, Units.W),
            "PV2 power": (9, Units.W),
            "On-grid total yield": (pack_u16(11, 12), Total(Units.KWH), div10),
            "On-grid daily yield": (13, DailyTotal(Units.KWH), div10),
            "Battery voltage": (14, Units.V, div100),
            "Battery current": (15, Units.A, div100),
            "Battery power": (16, Units.W),
            "Battery temperature": (17, Units.C),
            "Battery SoC": (18, Units.PERCENT),
            "Inverter Temperature": (26, Units.C),
            "Grid power": (32, Units.W, to_signed),
            "Total feed-in energy": (pack_u16(34, 35), Total(Units.KWH), div100),
            "Total consumption": (pack_u16(36, 37), Total(Units.KWH), div100),
        }

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]



import json
import logging
import sys
from collections import namedtuple
from typing import Any, Callable, Dict, Generator, Optional, Tuple, Union

import voluptuous as vol
from voluptuous import Invalid, MultipleInvalid
from voluptuous.humanize import humanize_error

from solax.units import SensorUnit
from solax.utils import PackerBuilderResult, contains_none_zero_value

__all__ = ("ResponseParser", "InverterResponse", "ResponseDecoder")

if sys.version_info >= (3, 11):
    from typing import Unpack
else:
    from typing_extensions import Unpack

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class InverterResponse(
    namedtuple(
        "InverterResponse",
        [
            "data",
            "dongle_serial_number",
            "version",
            "type",
            "inverter_serial_number",
        ],
    )
):
    @property
    def serial_number(self):
        return self.dongle_serial_number


_KEY_DATA = "data"
_KEY_SERIAL = "sn"
_KEY_VERSION = "version"
_KEY_VER = "ver"
_KEY_TYPE = "type"


GenericResponseSchema = vol.All(
    vol.Schema({vol.Required(_KEY_SERIAL): str}, extra=vol.ALLOW_EXTRA),
    vol.Any(
        vol.Schema({vol.Required(_KEY_VERSION): str}, extra=vol.ALLOW_EXTRA),
        vol.Schema({vol.Required(_KEY_VER): str}, extra=vol.ALLOW_EXTRA),
    ),
    vol.Schema(
        {
            vol.Required(_KEY_TYPE): vol.Any(int, str),
            vol.Required(_KEY_DATA): vol.Schema(contains_none_zero_value),
        },
        extra=vol.ALLOW_EXTRA,
    ),
)

ProcessorTuple = Tuple[Callable[[Any], Any], ...]
SensorIndexSpec = Union[int, PackerBuilderResult]
ResponseDecoder = Dict[
    str,
    Tuple[SensorIndexSpec, SensorUnit, Unpack[ProcessorTuple]],
]


class ResponseParser:
    def __init__(
        self,
        schema: vol.Schema,
        decoder: ResponseDecoder,
        dongle_serial_number_getter: Callable[[Dict[str, Any]], Optional[str]],
        inverter_serial_number_getter: Callable[[Dict[str, Any]], Optional[str]],
    ) -> None:
        self.schema = vol.And(GenericResponseSchema, schema)
        self.response_decoder = decoder
        self.dongle_serial_number_getter = dongle_serial_number_getter
        self.inverter_serial_number_getter = inverter_serial_number_getter

    def _decode_map(self) -> Dict[str, SensorIndexSpec]:
        sensors: Dict[str, SensorIndexSpec] = {}
        for name, mapping in self.response_decoder.items():
            sensors[name] = mapping[0]
        return sensors

    def _postprocess_gen(
        self,
    ) -> Generator[Tuple[str, Callable[[Any], Any]], None, None]:
        """
        Return map of functions to be applied to each sensor value
        """
        for name, mapping in self.response_decoder.items():
            (_, _, *processors) = mapping
            for processor in processors:
                yield name, processor

    def map_response(self, resp_data) -> Dict[str, Any]:
        result = {}
        for sensor_name, decode_info in self._decode_map().items():
            if isinstance(decode_info, (tuple, list)):
                indexes = decode_info[0]
                packer = decode_info[1]
                values = tuple(resp_data[i] for i in indexes)
                val = packer(*values)
            else:
                val = resp_data[decode_info]
            result[sensor_name] = val
        for sensor_name, processor in self._postprocess_gen():
            result[sensor_name] = processor(result[sensor_name])
        return result

    def handle_response(self, resp: bytearray) -> InverterResponse:
        """
        Decode response and map array result using mapping definition.

        Args:
            resp (bytearray): The response

        Returns:
            InverterResponse: The decoded and mapped interver response.
        """

        raw_json = resp.decode("utf-8").replace(",,", ",0.0,").replace(",,", ",0.0,")
        json_response = {}
        for key, value in json.loads(raw_json).items():
            json_response[key.lower()] = value

        try:
            response = self.schema(json_response)
        except (Invalid, MultipleInvalid) as ex:
            _ = humanize_error(json_response, ex)
            raise

        return InverterResponse(
            data=self.map_response(response[_KEY_DATA]),
            dongle_serial_number=self.dongle_serial_number_getter(response),
            version=response.get(_KEY_VER, response.get(_KEY_VERSION)),
            type=response[_KEY_TYPE],
            inverter_serial_number=self.inverter_serial_number_getter(response),
        )



import asyncio
from unittest.mock import Mock

import pytest

import solax


@pytest.mark.asyncio
async def test_waits_when_asked(monkeypatch):
    sleep_f = asyncio.Future()
    sleep_f.set_result(None)
    mock_sleep = Mock(return_value=sleep_f)
    monkeypatch.setattr(asyncio, "sleep", mock_sleep)

    inv = Mock()
    get_data_f = asyncio.Future()
    get_data_f.set_result({})
    inv.get_data = Mock(return_value=get_data_f)

    wait_time = 2
    await solax.rt_request(inv, 10, wait_time)

    mock_sleep.assert_called_once_with(wait_time)
    inv.get_data.assert_called_once()


@pytest.mark.asyncio
async def test_tries_again_on_timeout(monkeypatch):
    sleep_f = asyncio.Future()
    sleep_f.set_result(None)
    mock_sleep = Mock(return_value=sleep_f)
    monkeypatch.setattr(asyncio, "sleep", mock_sleep)

    inv = Mock()
    get_data_f = asyncio.Future()
    get_data_f.set_exception(asyncio.TimeoutError)
    inv.get_data = Mock(return_value=get_data_f)

    wait_time = 2
    with pytest.raises(asyncio.TimeoutError):
        await solax.rt_request(inv, 2, wait_time)

    assert mock_sleep.call_count == 2
    assert inv.get_data.call_count == 2



import struct

import pytest

from solax import utils


@pytest.mark.parametrize(
    "val_to_check",
    [
        0,
        1,
        0x7FFF,  # int16 max
        0x8000,  # - 2**15
        0xFFFF,  # -1
    ],
)
def test_to_signed(val_to_check):
    # Take input as an int so can bit-twiddle reliably
    actual = utils.to_signed(val_to_check)
    expected = struct.unpack("<h", struct.pack("<H", val_to_check))[0]
    assert actual == expected


@pytest.mark.parametrize(
    "val_to_check_32",
    [
        0,
        1,
        0x8000,  # 32768
        0xFFFF,  # 65535
        0x7FFFFFFF,  # int32 max
        0x80000000,  # - 2**31
        0x7FFFFFFF,  # -1
    ],
)
def test_to_signed32(val_to_check_32):
    # Take input as an int so can bit-twiddle reliably
    actual = utils.to_signed32(val_to_check_32)
    expected = struct.unpack("<i", struct.pack("<I", val_to_check_32))[0]
    assert actual == expected



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Measurement, Total, Units
from solax.utils import (
    div10,
    div100,
    pack_u16,
    to_signed,
    to_signed32,
    twoway_div10,
    twoway_div100,
)


class X3Ultra(Inverter):
    """X3 Ultra v1.001.20"""

    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(int, 25),
            vol.Required("sn"): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=300, max=300),
                )
            ),
            vol.Required("information"): vol.Schema(
                vol.All(vol.Length(min=10, max=10))
            ),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        return [cls._build(host, port, pwd, False)]

    @classmethod
    def _decode_run_mode(cls, run_mode):
        return {
            0: "Waiting",
            1: "Checking",
            2: "Normal",
            3: "Fault",
            4: "Permanent Fault",
            5: "Updating",
            6: "EPS Check",
            7: "EPS Mode",
            8: "Self Test",
            9: "Idle",
            10: "Standby",
        }.get(run_mode)

    @classmethod
    def response_decoder(cls):
        return {
            "Grid 1 Voltage": (0, Units.V, div10),
            "Grid 2 Voltage": (1, Units.V, div10),
            "Grid 3 Voltage": (2, Units.V, div10),
            "Grid 1 Current": (3, Units.A, twoway_div10),
            "Grid 2 Current": (4, Units.A, twoway_div10),
            "Grid 3 Current": (5, Units.A, twoway_div10),
            "Grid 1 Power": (6, Units.W, to_signed),
            "Grid 2 Power": (7, Units.W, to_signed),
            "Grid 3 Power": (8, Units.W, to_signed),
            "PV1 Voltage": (10, Units.V, div10),
            "PV2 Voltage": (11, Units.V, div10),
            "PV3 Voltage": (129, Units.V, div10),
            "PV1 Current": (12, Units.A, div10),
            "PV2 Current": (13, Units.A, div10),
            "PV3 Current": (130, Units.A, div10),
            "PV1 Power": (14, Units.W),
            "PV2 Power": (15, Units.W),
            "PV3 Power": (131, Units.W),
            "Grid 1 Frequency": (16, Units.HZ, div100),
            "Grid 2 Frequency": (17, Units.HZ, div100),
            "Grid 3 Frequency": (18, Units.HZ, div100),
            # "Run mode": (19, Units.NONE), # Only use the index once due to HA uids
            "Run mode text": (19, Units.NONE, X3Ultra._decode_run_mode),
            "EPS 1 Voltage": (23, Units.V, div10),
            "EPS 2 Voltage": (24, Units.V, div10),
            "EPS 3 Voltage": (25, Units.V, div10),
            "EPS 1 Current": (26, Units.A, twoway_div10),
            "EPS 2 Current": (27, Units.A, twoway_div10),
            "EPS 3 Current": (28, Units.A, twoway_div10),
            "EPS 1 Power": (29, Units.W, to_signed),
            "EPS 2 Power": (30, Units.W, to_signed),
            "EPS 3 Power": (31, Units.W, to_signed),
            "Grid Power ": (pack_u16(34, 35), Units.W, to_signed32),
            "Battery 1 Voltage": (39, Units.V, div10),
            "Battery 2 Voltage": (132, Units.V, div10),
            "Battery 1 Current": (40, Units.A, twoway_div100),
            "Battery 2 Current": (133, Units.A, twoway_div100),
            "Battery 1 Power": (41, Units.W, to_signed),
            "Battery 2 Power": (134, Units.W, to_signed),
            "Battery 1 Remaining Capacity": (103, Units.PERCENT),
            "Battery 2 Remaining Capacity": (140, Units.PERCENT),
            "Battery 1 Temperature": (105, Units.C, to_signed),
            "Battery 2 Temperature": (142, Units.C, to_signed),
            "Load/Generator Power": (47, Units.W, to_signed),
            "Radiator Temperature": (54, Units.C, to_signed),
            "Yield total": (pack_u16(58, 59), Total(Units.KWH), div10),
            "Yield today": (70, DailyTotal(Units.KWH), div10),
            "Battery Discharge Energy total": (
                pack_u16(74, 75),
                Total(Units.KWH),
                div10,
            ),
            "Battery Charge Energy total": (pack_u16(76, 77), Total(Units.KWH), div10),
            "Battery Discharge Energy today": (78, DailyTotal(Units.KWH), div10),
            "Battery Charge Energy today": (79, DailyTotal(Units.KWH), div10),
            "PV Energy total": (pack_u16(80, 81), Total(Units.KWH), div10),
            "EPS Energy total": (pack_u16(83, 84), Total(Units.KWH), div10),
            "EPS Energy today": (85, DailyTotal(Units.KWH), div10),
            "Feed-in Energy total": (pack_u16(86, 87), Total(Units.KWH), div100),
            "Grid Consumed Energy total": (pack_u16(88, 89), Total(Units.KWH), div100),
            "Feed-in Energy today": (pack_u16(90, 91), DailyTotal(Units.KWH), div100),
            "Grid Consumed Energy today": (
                pack_u16(92, 93),
                DailyTotal(Units.KWH),
                div100,
            ),
            "Battery Remaining Capacity": (158, Units.PERCENT),
            "Battery Remaining Energy": (
                106,
                Measurement(Units.KWH, storage=True),
                div10,
            ),
            "Inverter Power": (159, Units.W, div10),
        }

    # pylint: enable=duplicate-code

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]



""" Units and different measurement types"""

from enum import Enum
from typing import NamedTuple, Union


class Units(Enum):
    """All known Units."""

    W = "W"
    KWH = "kWh"
    A = "A"
    V = "V"
    C = "°C"
    HZ = "Hz"
    PERCENT = "%"
    NONE = ""


class Measurement(NamedTuple):
    """Representation of measurement with a given unit and arbitrary values."""

    unit: Units
    is_monotonic: bool = False
    resets_daily: bool = False
    storage: bool = False


class Total(Measurement):
    """A Measurement where the values are continuously increasing."""

    is_monotonic: bool = True


class DailyTotal(Measurement):
    """A Measurement where the values are reset daily."""

    resets_daily: bool = True


SensorUnit = Union[Measurement, Total]



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Total, Units
from solax.utils import div10, div100


class X1MiniV34(Inverter):
    """
    X1-Boost-Air-Mini with Wifi Pocket v2.034.06
    SolarX disabled lan access with this custom
    firmware you can access te Wifi Pocket from your internal lan:
    https://blog.chrisoft.io/2021/02/14/
    firmwares-modificados-para-solax-pocket-wifi-v2/
    """

    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type", "type"): vol.All(int, 4),
            vol.Required(
                "sn",
            ): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Any(
                        vol.Length(min=69, max=69),
                        vol.Length(min=100, max=100),
                        vol.Length(min=200, max=200),
                    ),
                )
            ),
            vol.Required("information"): vol.Schema(
                vol.Any(vol.Length(min=9, max=9), vol.Length(min=10, max=10))
            ),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def response_decoder(cls):
        return {
            "Network Voltage": (0, Units.V, div10),
            "Output Current": (1, Units.A, div10),
            "AC Power": (2, Units.W),
            "PV1 Voltage": (3, Units.V, div10),
            "PV2 Voltage": (4, Units.V, div10),
            "PV1 Current": (5, Units.A, div10),
            "PV2 Current": (6, Units.A, div10),
            "PV1 Power": (7, Units.W),
            "PV2 Power": (8, Units.W),
            "Grid Frequency": (9, Units.HZ, div100),
            "Total Energy": (11, Total(Units.KWH), div10),
            "Today's Energy": (13, DailyTotal(Units.KWH), div10),
            "Total Feed-in Energy": (41, Total(Units.KWH), div10),
            "Total Consumption": (42, Total(Units.KWH), div10),
            "Power Now": (43, Units.W, div10),
            "Inverter Temperature": (55, Units.C),
        }

    # pylint: enable=duplicate-code

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Total, Units
from solax.utils import div10, div100, to_signed


class X1Smart(Inverter):
    """
    X1-Smart with Pocket WiFi v2.033.20
    Includes X-Forwarded-For for direct LAN API access
    """

    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type", "type"): vol.All(int, 8),
            vol.Required(
                "sn",
            ): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=200, max=200),
                )
            ),
            vol.Required("information"): vol.Schema(vol.All(vol.Length(min=8, max=8))),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def response_decoder(cls):
        return {
            "Network Voltage": (0, Units.V, div10),
            "Output Current": (1, Units.A, div10),
            "AC Power": (2, Units.W),
            "PV1 Voltage": (3, Units.V, div10),
            "PV2 Voltage": (4, Units.V, div10),
            "PV1 Current": (5, Units.A, div10),
            "PV2 Current": (6, Units.A, div10),
            "PV1 Power": (7, Units.W),
            "PV2 Power": (8, Units.W),
            "Grid Frequency": (9, Units.HZ, div100),
            "Total Energy": (11, Total(Units.KWH), div10),
            "Today's Energy": (13, DailyTotal(Units.KWH), div10),
            "Inverter Temperature": (39, Units.C),
            "Exported Power": (48, Units.W, to_signed),
            "Total Feed-in Energy": (50, Total(Units.KWH), div100),
            "Total Consumption": (52, Total(Units.KWH), div100),
        }

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        versions = [
            cls._build(host, port, pwd, True),
            cls._build(host, port, pwd, False),
        ]
        for inverter in versions:
            inverter.http_client = inverter.http_client.with_headers(
                {"X-Forwarded-For": "5.8.8.8"}
            )
        return versions



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter, InverterHttpClient, Method
from solax.units import DailyTotal, Total, Units


class XHybrid(Inverter):
    """
    Tested with:
    * SK-TL5000E
    """

    _schema = vol.Schema(
        {
            vol.Required("method"): str,
            vol.Required("version"): str,
            vol.Required("type"): str,
            vol.Required("sn"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Any(vol.Length(min=58, max=58), vol.Length(min=68, max=68)),
                )
            ),
            vol.Required("status"): vol.All(vol.Coerce(int), vol.Range(min=0)),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def _build(cls, host, port, pwd="", params_in_query=True):
        base = "http://{}:{}/api/realTimeData.htm"
        url = base.format(host, port)
        http_client = InverterHttpClient(url=url, method=Method.GET, pwd="")

        return cls(http_client)

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        versions = [cls._build(host, port)]
        return versions

    # key: name of sensor
    # value.0: index
    # value.1: unit (String) or None
    # from https://github.com/GitHobi/solax/wiki/direct-data-retrieval
    @classmethod
    def response_decoder(cls):
        return {
            "PV1 Current": (0, Units.A),
            "PV2 Current": (1, Units.A),
            "PV1 Voltage": (2, Units.V),
            "PV2 Voltage": (3, Units.V),
            "Output Current": (4, Units.A),
            "Network Voltage": (5, Units.V),
            "Power Now": (6, Units.W),
            "Inverter Temperature": (7, Units.C),
            "Today's Energy": (8, DailyTotal(Units.KWH)),
            "Total Energy": (9, Total(Units.KWH)),
            "Exported Power": (10, Units.W),
            "PV1 Power": (11, Units.W),
            "PV2 Power": (12, Units.W),
            "Battery Voltage": (13, Units.V),
            "Battery Current": (14, Units.A),
            "Battery Power": (15, Units.W),
            "Battery Temperature": (16, Units.C),
            "Battery Remaining Capacity": (17, Units.PERCENT),
            "Month's Energy": (19, Units.KWH),
            "Grid Exported Energy": (41, Units.KWH),
            "Grid Imported Energy": (42, Units.KWH),
            "Grid Frequency": (50, Units.HZ),
            "EPS Voltage": (53, Units.V),
            "EPS Current": (54, Units.A),
            "EPS Power": (55, Units.W),
            "EPS Frequency": (56, Units.HZ),
        }

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return None



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Total, Units
from solax.utils import div10, div100, pack_u16, to_signed, to_signed32, twoway_div10


class X3MicProG2(Inverter):
    """X3MicProG2 v3.008.10"""

    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(int, 16),
            vol.Required("sn"): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=100, max=100),
                )
            ),
            vol.Required("information"): vol.Schema(
                vol.All(vol.Length(min=10, max=10))
            ),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        return [cls._build(host, port, pwd, False)]

    @classmethod
    def _decode_run_mode(cls, run_mode):
        return {
            0: "Wait",
            1: "Check",
            2: "Normal",
            3: "Fault",
            4: "Permanent Fault",
            5: "Update",
        }.get(run_mode)

    @classmethod
    def response_decoder(cls):
        return {
            "Grid 1 Voltage": (0, Units.V, div10),
            "Grid 2 Voltage": (1, Units.V, div10),
            "Grid 3 Voltage": (2, Units.V, div10),
            "Grid 1 Current": (3, Units.A, twoway_div10),
            "Grid 2 Current": (4, Units.A, twoway_div10),
            "Grid 3 Current": (5, Units.A, twoway_div10),
            "Grid 1 Power": (6, Units.W, to_signed),
            "Grid 2 Power": (7, Units.W, to_signed),
            "Grid 3 Power": (8, Units.W, to_signed),
            "PV1 Voltage": (9, Units.V, div10),
            "PV2 Voltage": (10, Units.V, div10),
            "PV3 Voltage": (11, Units.V, div10),
            "PV1 Current": (12, Units.A, div10),
            "PV2 Current": (13, Units.A, div10),
            "PV3 Current": (14, Units.A, div10),
            "PV1 Power": (15, Units.W),
            "PV2 Power": (16, Units.W),
            "PV3 Power": (17, Units.W),
            "Grid 1 Frequency": (18, Units.HZ, div100),
            "Grid 2 Frequency": (19, Units.HZ, div100),
            "Grid 3 Frequency": (20, Units.HZ, div100),
            # "Run Mode": (21, Units.NONE),
            "Run Mode": (21, Units.NONE, X3MicProG2._decode_run_mode),
            "Total Yield": (pack_u16(22, 23), Total(Units.KWH), div10),
            "Daily Yield": (24, DailyTotal(Units.KWH), div10),
            "Feed-in Power ": (pack_u16(72, 73), Units.W, to_signed32),
            "Total Feed-in Energy": (pack_u16(74, 75), Total(Units.KWH), div100),
            "Total Consumption": (pack_u16(76, 77), Total(Units.KWH), div100),
        }

    # pylint: enable=duplicate-code

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]



"""Support for Solax inverter via local API."""

import asyncio
import logging

from solax.discovery import discover
from solax.inverter import Inverter, InverterResponse
from solax.inverter_http_client import REQUEST_TIMEOUT

_LOGGER = logging.getLogger(__name__)

__all__ = (
    "discover",
    "real_time_api",
    "rt_request",
    "Inverter",
    "InverterResponse",
    "RealTimeAPI",
    "REQUEST_TIMEOUT",
)


async def rt_request(inv: Inverter, retry, t_wait=0) -> InverterResponse:
    """Make call to inverter endpoint."""
    if t_wait > 0:
        msg = "Timeout connecting to Solax inverter, waiting %d to retry."
        _LOGGER.error(msg, t_wait)
        await asyncio.sleep(t_wait)
    new_wait = (t_wait * 2) + 5
    retry = retry - 1
    try:
        return await inv.get_data()
    except asyncio.TimeoutError:
        if retry > 0:
            return await rt_request(inv, retry, new_wait)
        _LOGGER.error("Too many timeouts connecting to Solax.")
        raise


async def real_time_api(ip_address, port=80, pwd=""):
    i = await discover(ip_address, port, pwd, return_when=asyncio.FIRST_COMPLETED)
    return RealTimeAPI(i)


class RealTimeAPI:
    """Solax inverter real time API"""

    # pylint: disable=too-few-public-methods

    def __init__(self, inv: Inverter):
        """Initialize the API client."""
        self.inverter = inv

    async def get_data(self) -> InverterResponse:
        """Query the real time API"""
        return await rt_request(self.inverter, 3)



XHYBRID_VALUES = {
    "Today's Energy": 8.0,
    "Battery Current": 14.0,
    "Month's Energy": 19.0,
    "Battery Power": 15,
    "Battery Remaining Capacity": 17,
    "Battery Temperature": 16,
    "Battery Voltage": 13,
    "EPS Current": 54,
    "EPS Frequency": 56,
    "EPS Power": 55,
    "EPS Voltage": 53,
    "Exported Power": 10,
    "Grid Frequency": 50,
    "Inverter Temperature": 7,
    "Network Voltage": 5,
    "Output Current": 4,
    "PV1 Current": 0,
    "PV1 Power": 11,
    "PV1 Voltage": 2,
    "PV2 Current": 1,
    "PV2 Power": 12,
    "PV2 Voltage": 3,
    "Power Now": 6,
    "Total Energy": 9,
    "Grid Exported Energy": 41,
    "Grid Imported Energy": 42,
}


X3_VALUES = {
    "PV1 Current": 0,
    "PV2 Current": 1,
    "PV1 Voltage": 2,
    "PV2 Voltage": 3,
    "Output Current Phase 1": 4,
    "Network Voltage Phase 1": 5,
    "AC Power": 6,
    "Inverter Temperature": 7,
    "Today's Energy": 8,
    "Total Energy": 9,
    "Exported Power": 10,
    "PV1 Power": 11,
    "PV2 Power": 12,
    "Battery Voltage": 13,
    "Battery Current": 14,
    "Battery Power": 15,
    "Battery Temperature": 16,
    "Battery Remaining Capacity": 21,
    "Total Feed-in Energy": 41,
    "Total Consumption": 42,
    "Output Current Phase 2": 46,
    "Output Current Phase 3": 47,
    "Network Voltage Phase 2": 48,
    "Network Voltage Phase 3": 49,
    "Grid Frequency Phase 1": 50,
    "Grid Frequency Phase 2": 51,
    "Grid Frequency Phase 3": 52,
    "Power Now Phase 1": 43,
    "Power Now Phase 2": 44,
    "Power Now Phase 3": 45,
    "EPS Voltage": 53,
    "EPS Current": 54,
    "EPS Power": 55,
    "EPS Frequency": 56,
}

X3_HYBRID_VALUES = {
    "PV1 Current": 0,
    "PV2 Current": 0,
    "PV1 Voltage": 0,
    "PV2 Voltage": 0,
    "Output Current Phase 1": 0.9,
    "Network Voltage Phase 1": 234,
    "AC Power": 3189,
    "Inverter Temperature": 42,
    "Today's Energy": 15.2,
    "Total Energy": 27,
    "Exported Power": -25,
    "PV1 Power": 0,
    "PV2 Power": 0,
    "Battery Voltage": 210.3,
    "Battery Current": -15.7,
    "Battery Power": -3321,
    "Battery Temperature": 24,
    "Battery Remaining Capacity": 68,
    "Total Feed-in Energy": 2.03,
    "Total Consumption": 23.41,
    "Output Current Phase 2": 5.4,
    "Output Current Phase 3": 6.9,
    "Network Voltage Phase 2": 250,
    "Network Voltage Phase 3": 251.9,
    "Grid Frequency Phase 1": 50.01,
    "Grid Frequency Phase 2": 50.01,
    "Grid Frequency Phase 3": 50.01,
    "Power Now Phase 1": 123,
    "Power Now Phase 2": 1344,
    "Power Now Phase 3": 1722,
    "EPS Voltage": 0,
    "EPS Current": 0,
    "EPS Power": 0,
    "EPS Frequency": 0,
}

X3V34_HYBRID_VALUES = {
    "Network Voltage Phase 1": 246.8,
    "Network Voltage Phase 2": 249,
    "Network Voltage Phase 3": 250.8,
    "Output Current Phase 1": 1.3,
    "Output Current Phase 2": 1.4,
    "Output Current Phase 3": 1,
    "Power Now Phase 1": 266,
    "Power Now Phase 2": 284,
    "Power Now Phase 3": 136,
    "PV1 Voltage": 537.7,
    "PV2 Voltage": 463,
    "PV1 Current": 1.7,
    "PV2 Current": 0,
    "PV1 Power": 958,
    "PV2 Power": 0,
    "Total PV Energy": 1731.9,
    "Today's PV Energy": 16.4,
    "Grid Frequency Phase 1": 50.03,
    "Grid Frequency Phase 2": 50.03,
    "Grid Frequency Phase 3": 50.03,
    "Total Energy": 1483.3,
    "Today's Energy": 10.3,
    "Battery Voltage": 229.3,
    "Battery Current": 0.9,
    "Battery Power": 229,
    "Battery Temperature": 22,
    "Battery Remaining Capacity": 99,
    "Total Battery Discharge Energy": 706.2,
    "Today's Battery Discharge Energy": 4.3,
    "Battery Remaining Energy": 12.5,
    "Total Battery Charge Energy": 814.2,
    "Today's Battery Charge Energy": 9.1,
    "Exported Power": -56,
    "Total Feed-in Energy": 173.72,
    "Total Consumption": 598.77,
    "AC Power": 686,
    "EPS Frequency": 0,
    "EPS Total Energy": 0.6,
}

X3V34_HYBRID_VALUES_NEGATIVE_POWER = {
    "Network Voltage Phase 1": 236.4,
    "Network Voltage Phase 2": 243.1,
    "Network Voltage Phase 3": 238.6,
    "Output Current Phase 1": -7.3,
    "Output Current Phase 2": -7.2,
    "Output Current Phase 3": -7.2,
    "Power Now Phase 1": -1738,
    "Power Now Phase 2": -1767,
    "Power Now Phase 3": -1729,
    "PV1 Voltage": 0,
    "PV2 Voltage": 0,
    "PV1 Current": 0,
    "PV2 Current": 0,
    "PV1 Power": 0,
    "PV2 Power": 0,
    "Total PV Energy": 9615.2,
    "Today's PV Energy": 9.8,
    "Grid Frequency Phase 1": 49.98,
    "Grid Frequency Phase 2": 49.98,
    "Grid Frequency Phase 3": 49.98,
    "Total Energy": 8478.7,
    "Today's Energy": 8.4,
    "Battery Voltage": 204.6,
    "Battery Current": 25,
    "Battery Power": 5117,
    "Battery Temperature": 24,
    "Battery Remaining Capacity": 20,
    "Total Battery Discharge Energy": 2469.6,
    "Today's Battery Discharge Energy": 4,
    "Battery Remaining Energy": 2.6,
    "Total Battery Charge Energy": 2887.6,
    "Today's Battery Charge Energy": 6,
    "Exported Power": -5744,
    "Total Feed-in Energy": 3593.54,
    "Total Consumption": 945.63,
    "AC Power": -5234,
    "EPS Frequency": 0,
    "EPS Total Energy": 2.1,
}

X3V34_HYBRID_VALUES_EPS_MODE = {
    "Network Voltage Phase 1": 0,
    "Network Voltage Phase 2": 0,
    "Network Voltage Phase 3": 0,
    "Output Current Phase 1": 0,
    "Output Current Phase 2": 0,
    "Output Current Phase 3": 0,
    "Power Now Phase 1": 0,
    "Power Now Phase 2": 0,
    "Power Now Phase 3": 0,
    "PV1 Voltage": 486.4,
    "PV2 Voltage": 446.6,
    "PV1 Current": 5.5,
    "PV2 Current": 0,
    "PV1 Power": 2678,
    "PV2 Power": 0,
    "Total PV Energy": 8378.8,
    "Today's PV Energy": 26,
    "Grid Frequency Phase 1": 0,
    "Grid Frequency Phase 2": 0,
    "Grid Frequency Phase 3": 0,
    "Total Energy": 7387.5,
    "Today's Energy": 17.8,
    "Battery Voltage": 228.6,
    "Battery Current": -1.00,
    "Battery Power": -242,
    "Battery Temperature": 27,
    "Battery Remaining Capacity": 98,
    "Total Battery Discharge Energy": 2124,
    "Today's Battery Discharge Energy": 6.1,
    "Battery Remaining Energy": 12.4,
    "Total Battery Charge Energy": 2501.1,
    "Today's Battery Charge Energy": 11.9,
    "Exported Power": 0,
    "Total Feed-in Energy": 3174.83,
    "Total Consumption": 844.58,
    "AC Power": 0,
    "EPS Frequency": 50,
    "EPS Total Energy": 1.7,
}

X3_HYBRID_G4_VALUES = {
    "Grid 1 Voltage": 238.3,
    "Grid 2 Voltage": 239.7,
    "Grid 3 Voltage": 237.7,
    "Grid 1 Current": 1.0,
    "Grid 2 Current": 1.1,
    "Grid 3 Current": 1.1,
    "Grid 1 Power": 33.0,
    "Grid 2 Power": 81.0,
    "Grid 3 Power": 76.0,
    "PV1 Voltage": 248.3,
    "PV2 Voltage": 230.5,
    "PV1 Current": 0.1,
    "PV2 Current": 0.2,
    "PV1 Power": 45.0,
    "PV2 Power": 58.0,
    "Grid 1 Frequency": 50.01,
    "Grid 2 Frequency": 50.01,
    "Grid 3 Frequency": 50.01,
    "Run mode text": "Normal",
    "EPS 1 Voltage": 0.0,
    "EPS 2 Voltage": 0.0,
    "EPS 3 Voltage": 0.0,
    "EPS 1 Current": 0.0,
    "EPS 2 Current": 0.0,
    "EPS 3 Current": 0.0,
    "EPS 1 Power": 0.0,
    "EPS 2 Power": 0.0,
    "EPS 3 Power": 0.0,
    "Grid Power ": -152.0,
    "Battery Current": 0.0,
    "Battery Power": 0.0,
    "Load/Generator Power": 342.0,
    "Radiator Temperature": 23.0,
    "Yield total": 4.4,
    "Yield today": 0.1,
    "Battery Discharge Energy total": 2.4,
    "Battery Charge Energy total": 1.0,
    "Battery Discharge Energy today": 0.0,
    "Battery Charge Energy today": 0.0,
    "PV Energy total": 3.4,
    "EPS Energy total": 0.0,
    "EPS Energy today": 0.0,
    "Feed-in Energy": 0.0,
    "Consumed Energy": 20.09,
    "Feed-in Energy total": 0.0,
    "Consumed Energy total": 2.09,
    "Battery Remaining Capacity": 30.0,
    "Battery Temperature": 22.0,
    "Battery Remaining Energy": 3.3,
    "Battery mode": 0.0,
    "Battery mode text": "Self Use Mode",
    "Battery Voltage": 234.6,
}

X3_MICPRO_G2_VALUES = {
    "Grid 1 Voltage": 237.1,
    "Grid 2 Voltage": 234.5,
    "Grid 3 Voltage": 237.8,
    "Grid 1 Current": 4.1,
    "Grid 2 Current": 4.1,
    "Grid 3 Current": 4.1,
    "Grid 1 Power": 1018.0,
    "Grid 2 Power": 992.0,
    "Grid 3 Power": 970.0,
    "PV1 Voltage": 174.1,
    "PV2 Voltage": 174.5,
    "PV3 Voltage": 0.0,
    "PV1 Current": 8.8,
    "PV2 Current": 8.7,
    "PV3 Current": 0.0,
    "PV1 Power": 1544.0,
    "PV2 Power": 1531.0,
    "PV3 Power": 0.0,
    "Grid 1 Frequency": 49.98,
    "Grid 2 Frequency": 49.99,
    "Grid 3 Frequency": 49.94,
    "Run Mode": "Normal",
    "Total Yield": 795.7,
    "Daily Yield": 20.0,
    "Feed-in Power ": 2707.0,
    "Total Feed-in Energy": 657.24,
    "Total Consumption": 307.33,
}

X1_VALUES = {
    "PV1 Current": 0,
    "PV2 Current": 1,
    "PV1 Voltage": 2,
    "PV2 Voltage": 3,
    "Output Current": 4,
    "Network Voltage": 5,
    "AC Power": 6,
    "Inverter Temperature": 7,
    "Today's Energy": 8,
    "Total Energy": 9,
    "Exported Power": 10,
    "PV1 Power": 11,
    "PV2 Power": 12,
    "Battery Voltage": 13,
    "Battery Current": 14,
    "Battery Power": 15,
    "Battery Temperature": 16,
    "Battery Remaining Capacity": 21,
    "Total Feed-in Energy": 41,
    "Total Consumption": 42,
    "Grid Frequency": 50,
    "Power Now": 43,
    "EPS Voltage": 53,
    "EPS Current": 54,
    "EPS Power": 55,
    "EPS Frequency": 56,
}

X1_MINI_VALUES = {
    "PV1 Current": 0,
    "PV2 Current": 1,
    "PV1 Voltage": 2,
    "PV2 Voltage": 3,
    "Output Current": 4,
    "Network Voltage": 5,
    "AC Power": 6,
    "Inverter Temperature": 7,
    "Today's Energy": 8,
    "Total Energy": 9,
    "Exported Power": 10,
    "PV1 Power": 11,
    "PV2 Power": 12,
    "Total Feed-in Energy": 41,
    "Total Consumption": 42,
    "Power Now": 43,
    "Grid Frequency": 50,
}

X1_MINI_VALUES_V34 = {
    "Network Voltage": 231.0,
    "Output Current": 1.5,
    "AC Power": 349,
    "PV1 Voltage": 60.9,
    "PV2 Voltage": 0,
    "PV1 Current": 5.8,
    "PV2 Current": 0,
    "PV1 Power": 359,
    "PV2 Power": 0,
    "Grid Frequency": 50.0,
    "Total Energy": 347.4,
    "Today's Energy": 1.2,
    "Total Feed-in Energy": 251.8,
    "Total Consumption": 0,
    "Power Now": 0,
    "Inverter Temperature": 41,
}


X1_MINI_VALUES_V34_VER3 = {
    "Network Voltage": 234.8,
    "Output Current": 0.7,
    "AC Power": 144.0,
    "PV1 Voltage": 96.6,
    "PV2 Voltage": 0.0,
    "PV1 Current": 1.6,
    "PV2 Current": 0.0,
    "PV1 Power": 154.0,
    "PV2 Power": 0.0,
    "Grid Frequency": 49.94,
    "Total Energy": 6.7,
    "Today's Energy": 3.1,
    "Total Feed-in Energy": 2.5,
    "Total Consumption": 0.0,
    "Power Now": 0.0,
    "Inverter Temperature": 39.0,
}

X1_SMART_VALUES = {
    "Network Voltage": 239.6,
    "Output Current": 12.6,
    "AC Power": 2956,
    "PV1 Voltage": 432.9,
    "PV2 Voltage": 233.8,
    "PV1 Current": 4.1,
    "PV2 Current": 5.7,
    "PV1 Power": 1777,
    "PV2 Power": 1336,
    "Grid Frequency": 60.21,
    "Total Energy": 121.5,
    "Today's Energy": 13.7,
    "Inverter Temperature": 43,
    "Exported Power": 10,
    "Total Feed-in Energy": 0.87,
    "Total Consumption": 81.84,
}

X1_BOOST_VALUES = {
    "AC Voltage": 239.6,
    "AC Output Current": 12.6,
    "AC Output Power": 2956,
    "PV1 Voltage": 432.9,
    "PV2 Voltage": 233.8,
    "PV1 Current": 4.1,
    "PV2 Current": 5.7,
    "PV1 Power": 1777,
    "PV2 Power": 1336,
    "AC Frequency": 60.21,
    "Total Generated Energy": 121.5,
    "Today's Generated Energy": 13.7,
    "Inverter Temperature": 43,
    "Exported Power": 10,
    "Total Export Energy": 0.87,
    "Total Import Energy": 81.84,
}

X1_BOOST_VALUES_OVERFLOWN = {
    "AC Voltage": 237.4,
    "AC Output Current": 6.7,
    "AC Output Power": 1581,
    "PV1 Voltage": 397.3,
    "PV2 Voltage": 233.8,
    "PV1 Current": 4.1,
    "PV2 Current": 5.7,
    "PV1 Power": 1628,
    "PV2 Power": 1336,
    "AC Frequency": 49.96,
    "Total Generated Energy": 14055.6,
    "Today's Generated Energy": 1.3,
    "Inverter Temperature": 36,
    "Exported Power": 1444,
    "Total Export Energy": 11348.18,
    "Total Import Energy": 1850.05,
}

QVOLTHYBG33P_VALUES = {
    "Network Voltage Phase 1": 221.4,
    "Network Voltage Phase 2": 223.8,
    "Network Voltage Phase 3": 225.1,
    "Output Current Phase 1": 1.1,
    "Output Current Phase 2": 1.0,
    "Output Current Phase 3": 1.2,
    "Power Now Phase 1": 162.0,
    "Power Now Phase 2": 136.0,
    "Power Now Phase 3": 146.0,
    "AC Power": 444.0,
    "PV1 Voltage": 566.2,
    "PV2 Voltage": 568.2,
    "PV1 Current": 1.8,
    "PV2 Current": 1.7,
    "PV1 Power": 1050.0,
    "PV2 Power": 977.0,
    "Grid Frequency Phase 1": 50.02,
    "Grid Frequency Phase 2": 50.01,
    "Grid Frequency Phase 3": 50.02,
    "Inverter Operation mode": "Normal",
    "Exported Power": -7.0,
    "Battery Voltage": 323.4,
    "Battery Current": 5.0,
    "Battery Power": 1616.0,
    "Power Now": 451.0,
    "Total Energy": 219.0,
    "Total Battery Discharge Energy": 73.8,
    "Total Battery Charge Energy": 90.4,
    "Today's Battery Discharge Energy": 0.0,
    "Today's Battery Charge Energy": 8.1,
    "Total PV Energy": 231.6,
    "Today's Energy": 11.8,
    "Total Feed-in Energy": 107.94,
    "Total Consumption": 145.44,
    "Today's Feed-in Energy": 1.66,
    "Today's Consumption": 4.66,
    "Battery Remaining Capacity": 95.0,
    "Battery Temperature": 35.0,
    "Battery Remaining Energy": 8.8,
    "Battery Operation mode": "Self Use Mode",
}

X1_HYBRID_G4_VALUES = {
    "AC voltage R": 247.0,
    "AC current": 1.1,
    "AC power": 237.0,
    "Grid frequency": 49.86,
    "PV1 voltage": 254.4,
    "PV2 voltage": 259.1,
    "PV1 current": 1.1,
    "PV2 current": 1.2,
    "PV1 power": 299.0,
    "PV2 power": 329.0,
    "On-grid total yield": 398.6,
    "On-grid daily yield": 2.2,
    "Battery voltage": 239.30,
    "Battery current": 2.20,
    "Battery power": 541.0,
    "Battery temperature": 26.0,
    "Battery SoC": 82.0,
    "Inverter Temperature": 30.0,
    "Grid power": 1.0,
    "Total feed-in energy": 286.7,
    "Total consumption": 6.2,
}

X3_ULTRA_VALUES = {
    "Grid 1 Voltage": 232.8,
    "Grid 2 Voltage": 233.8,
    "Grid 3 Voltage": 232.0,
    "Grid 1 Current": 2.0,
    "Grid 2 Current": 1.9,
    "Grid 3 Current": 1.9,
    "Grid 1 Power": 216.0,
    "Grid 2 Power": 249.0,
    "Grid 3 Power": 199.0,
    "PV1 Voltage": 0.0,
    "PV2 Voltage": 249.8,
    "PV3 Voltage": 245.4,
    "PV1 Current": 0.0,
    "PV2 Current": 6.5,
    "PV3 Current": 10.5,
    "PV1 Power": 0.0,
    "PV2 Power": 1629.0,
    "PV3 Power": 2592.0,
    "Grid 1 Frequency": 50.0,
    "Grid 2 Frequency": 50.0,
    "Grid 3 Frequency": 50.0,
    "Run mode text": "Normal",
    "EPS 1 Voltage": 0.0,
    "EPS 2 Voltage": 0.0,
    "EPS 3 Voltage": 0.0,
    "EPS 1 Current": 0.0,
    "EPS 2 Current": 0.0,
    "EPS 3 Current": 0.0,
    "EPS 1 Power": 0.0,
    "EPS 2 Power": 0.0,
    "EPS 3 Power": 0.0,
    "Grid Power ": 52.0,
    "Battery 1 Voltage": 213.8,
    "Battery 2 Voltage": 212.9,
    "Battery 1 Current": 9.6,
    "Battery 2 Current": 7.4,
    "Battery 1 Power": 2057.0,
    "Battery 2 Power": 1595.0,
    "Battery 1 Remaining Capacity": 28.0,
    "Battery 2 Remaining Capacity": 28.0,
    "Battery 1 Temperature": 13.0,
    "Battery 2 Temperature": 12.0,
    "Load/Generator Power": -52.0,
    "Radiator Temperature": 34.0,
    "Yield total": 198.4,
    "Yield today": 2.4,
    "Battery Discharge Energy total": 109.2,
    "Battery Charge Energy total": 122.8,
    "Battery Discharge Energy today": 0.2,
    "Battery Charge Energy today": 2.7,
    "PV Energy total": 122.9,
    "EPS Energy total": 0.0,
    "EPS Energy today": 0.0,
    "Feed-in Energy total": 27.71,
    "Grid Consumed Energy total": 128.12,
    "Feed-in Energy today": 0.1,
    "Grid Consumed Energy today": 8.6,
    "Battery Remaining Capacity": 28.0,
    "Battery Remaining Energy": 1.7,
    "Inverter Power": 66.4,
}



from numbers import Number
from typing import List, Protocol, Tuple

from voluptuous import Invalid


class Packer(Protocol):  # pragma: no cover
    # pylint: disable=R0903
    """
    Pack multiple raw values from the inverter
     data into one raw value
    """

    def __call__(self, *vals: float) -> float: ...


PackerBuilderResult = Tuple[Tuple[int, ...], Packer]


class PackerBuilder(Protocol):  # pragma: no cover
    # pylint: disable=R0903
    """
    Build a packer by identifying the indexes of the
    raw values to be fed to the packer
    """

    def __call__(self, *indexes: int) -> PackerBuilderResult: ...


def __u16_packer(*values: float) -> float:
    accumulator = 0.0
    stride = 1
    for value in values:
        accumulator += value * stride
        stride *= 2**16
    return accumulator


def pack_u16(*indexes: int) -> PackerBuilderResult:
    """
    Some values are expressed over 2 (or potentially
    more 16 bit [aka "short"] registers). Here we combine
    them, in order of least to most significant.
    """
    return (indexes, __u16_packer)


def startswith(something):
    def inner(actual):
        if isinstance(actual, str):
            if actual.startswith(something):
                return actual
        raise Invalid(f"{str(actual)} does not start with {something}")

    return inner


def div10(val):
    return val / 10


def div100(val):
    return val / 100


INT16_MAX = 0x7FFF
INT32_MAX = 0x7FFFFFFF


def to_signed(val):
    if val > INT16_MAX:
        val -= 2**16
    return val


def to_signed32(val):
    if val > INT32_MAX:
        val -= 2**32
    return val


def twoway_div10(val):
    return to_signed(val) / 10


def twoway_div100(val):
    return to_signed(val) / 100


def to_url(host, port):
    return f"http://{host}:{port}/"


def contains_none_zero_value(value: List[Number]):
    """Validate that at least one element is not zero.
    Args:
        value (List[Number]): list to validate
    Raises:
        Invalid: if all elements are zero
    """

    if isinstance(value, list):
        if len(value) != 0 and any((v != 0 for v in value)):
            return value
    raise Invalid("All elements in the list are zero")



import asyncio
import logging
import sys
from asyncio import Future, Task
from collections import defaultdict
from typing import Dict, Literal, Sequence, Set, Type, TypedDict, Union, cast

from solax.inverter import Inverter
from solax.inverter_http_client import InverterHttpClient

__all__ = ("discover", "DiscoveryKeywords", "DiscoveryError")

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points

if sys.version_info >= (3, 11):
    from typing import Unpack
else:
    from typing_extensions import Unpack

# registry of inverters
REGISTRY: Set[Type[Inverter]] = {
    ep.load()
    for ep in entry_points(group="solax.inverter")
    if issubclass(ep.load(), Inverter)
}

logging.basicConfig(level=logging.INFO)


class DiscoveryKeywords(TypedDict, total=False):
    inverters: Sequence[Type[Inverter]]
    return_when: Literal["ALL_COMPLETED", "FIRST_COMPLETED"]


if sys.version_info >= (3, 9):
    _InverterTask = Task[Inverter]
else:
    _InverterTask = Task


class _DiscoveryHttpClient:
    def __init__(
        self,
        inverter: Inverter,
        http_client: InverterHttpClient,
        request: Future,
    ):
        self._inverter = inverter
        self._http_client = http_client
        self._request: Future = request

    def __str__(self):
        return str(self._http_client)

    async def request(self):
        request = await self._request
        request.add_done_callback(self._restore_http_client)
        return await request

    def _restore_http_client(self, _: _InverterTask):
        self._inverter.http_client = self._http_client


async def _discovery_task(i) -> Inverter:
    logging.info("Trying inverter %s", i)
    await i.get_data()
    return i


async def discover(
    host, port, pwd="", **kwargs: Unpack[DiscoveryKeywords]
) -> Union[Inverter, Set[Inverter]]:
    done: Set[_InverterTask] = set()
    pending: Set[_InverterTask] = set()
    failures = set()
    requests: Dict[InverterHttpClient, Future] = defaultdict(
        asyncio.get_running_loop().create_future
    )

    return_when = kwargs.get("return_when", asyncio.FIRST_COMPLETED)
    for cls in kwargs.get("inverters", REGISTRY):
        for inverter in cls.build_all_variants(host, port, pwd):
            inverter.http_client = cast(
                InverterHttpClient,
                _DiscoveryHttpClient(
                    inverter, inverter.http_client, requests[inverter.http_client]
                ),
            )

            pending.add(
                asyncio.create_task(_discovery_task(inverter), name=f"{inverter}")
            )

    if not pending:
        raise DiscoveryError("No inverters to try to discover")

    def cancel(pending: Set[_InverterTask]) -> Set[_InverterTask]:
        for task in pending:
            task.cancel()
        return pending

    def remove_failures_from(done: Set[_InverterTask]) -> None:
        for task in set(done):
            exc = task.exception()
            if exc:
                failures.add(exc)
                done.remove(task)

    # stagger HTTP request to prevent accidental Denial Of Service
    async def stagger() -> None:
        for http_client, future in requests.items():
            future.set_result(asyncio.create_task(http_client.request()))
            await asyncio.sleep(1)

    staggered = asyncio.create_task(stagger())

    while pending and (not done or return_when != asyncio.FIRST_COMPLETED):
        try:
            done, pending = await asyncio.wait(pending, return_when=return_when)
        except asyncio.CancelledError:
            staggered.cancel()
            await asyncio.gather(staggered, *cancel(pending), return_exceptions=True)
            raise

        remove_failures_from(done)

        if done and return_when == asyncio.FIRST_COMPLETED:
            break

        logging.debug("%d discovery tasks are still running...", len(pending))

        if pending and return_when != asyncio.FIRST_COMPLETED:
            pending.update(done)
            done.clear()

    remove_failures_from(done)
    staggered.cancel()
    await asyncio.gather(staggered, *cancel(pending), return_exceptions=True)

    if done:
        logging.info("Discovered inverters: %s", {task.result() for task in done})
        if return_when == asyncio.FIRST_COMPLETED:
            return await next(iter(done))

        return {task.result() for task in done}

    raise DiscoveryError(
        "Unable to connect to the inverter at "
        f"host={host} port={port}, or your inverter is not supported yet.\n"
        "Please see https://github.com/squishykid/solax/wiki/DiscoveryError\n"
        f"Failures={str(failures)}"
    )


class DiscoveryError(Exception):
    """Raised when unable to discover inverter"""



import pytest

from solax.discovery import REGISTRY
from solax.inverter import Inverter


def test_all_registered_inverters_inherit_from_base():
    assert REGISTRY
    for i in REGISTRY:
        assert issubclass(i, Inverter)


def test_unimplemented_response_decoder():
    with pytest.raises(NotImplementedError):
        versions = Inverter.build_all_variants("localhost", 80)
        next(iter(versions)).response_decoder()



import asyncio

import pytest

import solax
from solax import InverterResponse
from solax.discovery import REGISTRY, DiscoveryError
from solax.inverter import InverterError
from solax.inverters import X1Boost


class DelayedX1Boost(X1Boost):
    async def get_data(self) -> InverterResponse:
        await asyncio.sleep(10)
        return await super().get_data()


class DelayedFailedX1Boost(X1Boost):
    async def make_request(self) -> InverterResponse:
        await asyncio.sleep(5)
        raise InverterError


@pytest.mark.asyncio
async def test_discovery(inverters_fixture):
    conn, inverter_class, _ = inverters_fixture
    inverters = await solax.discover(*conn, return_when=asyncio.ALL_COMPLETED)
    assert inverter_class in {type(inverter) for inverter in inverters}

    for inverter in inverters:
        if isinstance(inverter, inverter_class):
            data = await inverter.get_data()
            assert "X" * 7 in (data.inverter_serial_number or "X" * 7)
            assert data.serial_number == data.dongle_serial_number


@pytest.mark.asyncio
async def test_real_time_api(inverters_fixture):
    conn, inverter_class, _ = inverters_fixture

    if inverter_class is not X1Boost:
        pytest.skip()

    rt_api = await solax.real_time_api(*conn)
    assert rt_api.inverter.__class__ is inverter_class


@pytest.mark.asyncio
async def test_discovery_cancelled_error_while_staggering(
    inverters_fixture,
):
    conn, inverter_class, _ = inverters_fixture

    if inverter_class is not X1Boost:
        pytest.skip()

    task = asyncio.create_task(
        solax.discover(*conn, return_when=asyncio.FIRST_EXCEPTION)
    )
    await asyncio.sleep(1)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task


@pytest.mark.asyncio
async def test_discovery_cancelled_error_after_staggering(
    inverters_fixture,
):
    conn, inverter_class, _ = inverters_fixture

    if inverter_class is not X1Boost:
        pytest.skip()

    inverters = set(REGISTRY)
    inverters.add(DelayedX1Boost)

    task = asyncio.create_task(
        solax.discover(*conn, inverters=inverters, return_when=asyncio.FIRST_EXCEPTION)
    )
    await asyncio.sleep(7)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task


@pytest.mark.asyncio
async def test_discovery_first_completed_after_staggering(
    inverters_fixture,
):
    conn, inverter_class, _ = inverters_fixture

    if inverter_class is not X1Boost:
        pytest.skip()

    inverter = await solax.discover(
        *conn, inverters=[DelayedX1Boost], return_when=asyncio.FIRST_COMPLETED
    )
    assert inverter.__class__ is DelayedX1Boost


@pytest.mark.asyncio
async def test_discovery_not_first_completed_after_staggering(
    inverters_fixture,
):
    conn, inverter_class, _ = inverters_fixture

    if inverter_class is not X1Boost:
        pytest.skip()

    inverters = await solax.discover(
        *conn,
        inverters=[DelayedX1Boost, DelayedFailedX1Boost],
        return_when=asyncio.FIRST_EXCEPTION
    )
    assert DelayedX1Boost in {type(inverter) for inverter in inverters}


@pytest.mark.asyncio
async def test_discovery_no_host():
    with pytest.raises(DiscoveryError):
        await solax.real_time_api("localhost", 2)


@pytest.mark.asyncio
async def test_discovery_no_host_with_pwd():
    with pytest.raises(DiscoveryError):
        await solax.real_time_api("localhost", 2, "pwd")


@pytest.mark.asyncio
async def test_discovery_unknown_webserver(simple_http_fixture):
    with pytest.raises(DiscoveryError):
        await solax.real_time_api(*simple_http_fixture)


@pytest.mark.asyncio
async def test_discovery_empty_inverter_class_iterable():
    with pytest.raises(DiscoveryError):
        await solax.discover("localhost", 2, inverters=[])



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter, InverterHttpClient
from solax.units import DailyTotal, Measurement, Total, Units
from solax.utils import div10, div100, pack_u16, to_signed, twoway_div10, twoway_div100


class QVOLTHYBG33P(Inverter):
    """
    QCells
    Q.VOLT HYB-G3-3P
    """

    class Processors:
        """
        Postprocessors used only in the QVOLTHYBG33P inverter sensor_map.
        """

        @staticmethod
        def inverter_modes(value):
            return {
                0: "Waiting",
                1: "Checking",
                2: "Normal",
                3: "Off",
                4: "Permanent Fault",
                5: "Updating",
                6: "EPS Check",
                7: "EPS Mode",
                8: "Self Test",
                9: "Idle",
                10: "Standby",
            }.get(value, f"unmapped value '{value}'")

        @staticmethod
        def battery_modes(value):
            return {
                0: "Self Use Mode",
                1: "Force Time Use",
                2: "Back Up Mode",
                3: "Feed-in Priority",
            }.get(value, f"unmapped value '{value}'")

    def __init__(self, http_client: InverterHttpClient, *args, **kwargs):
        super().__init__(http_client, *args, **kwargs)
        self.manufacturer = "Qcells"

    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(int, 14),
            vol.Required("sn"): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Length(min=200, max=200),
                )
            ),
            vol.Required("information"): vol.Schema(
                vol.All(vol.Length(min=10, max=10))
            ),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def response_decoder(cls):
        return {
            "Network Voltage Phase 1": (0, Units.V, div10),
            "Network Voltage Phase 2": (1, Units.V, div10),
            "Network Voltage Phase 3": (2, Units.V, div10),
            "Output Current Phase 1": (3, Units.A, twoway_div10),
            "Output Current Phase 2": (4, Units.A, twoway_div10),
            "Output Current Phase 3": (5, Units.A, twoway_div10),
            "Power Now Phase 1": (6, Units.W, to_signed),
            "Power Now Phase 2": (7, Units.W, to_signed),
            "Power Now Phase 3": (8, Units.W, to_signed),
            "AC Power": (9, Units.W, to_signed),
            "PV1 Voltage": (10, Units.V, div10),
            "PV2 Voltage": (11, Units.V, div10),
            "PV1 Current": (12, Units.A, div10),
            "PV2 Current": (13, Units.A, div10),
            "PV1 Power": (14, Units.W),
            "PV2 Power": (15, Units.W),
            "Grid Frequency Phase 1": (16, Units.HZ, div100),
            "Grid Frequency Phase 2": (17, Units.HZ, div100),
            "Grid Frequency Phase 3": (18, Units.HZ, div100),
            "Inverter Operation mode": (19, Units.NONE, cls.Processors.inverter_modes),
            # 20 - 32: always 0
            # 33: always 1
            # instead of to_signed this is actually 34 - 35,
            # because 35 =  if 34>32767: 0 else: 65535
            "Exported Power": (34, Units.W, to_signed),
            # 35: if 34>32767: 0 else: 65535
            # 36 - 38    : always  0
            "Battery Voltage": (39, Units.V, div100),
            "Battery Current": (40, Units.A, twoway_div100),
            "Battery Power": (41, Units.W, to_signed),
            # 42: div10, almost identical to [39]
            # 43: twoway_div10,  almost the same as "40" (battery current)
            # 44: twoway_div100, almost the same as "41" (battery power),
            # 45: always 1
            # 46: follows PV Output, idles around 44, peaks at 52,
            "Power Now": (47, Units.W, to_signed),
            # 48: always 256
            # 49,50: [49] + [50] * 15160 some increasing counter
            # 51: always 5634
            # 52: always 100
            # 53: always 0
            # 54: follows PV Output, idles around 35, peaks at 54,
            # 55-67: always 0
            "Total Energy": (pack_u16(68, 69), Total(Units.KWH), div10),
            # 70: div10, today's energy including battery usage
            # 71-73: 0
            "Total Battery Discharge Energy": (
                pack_u16(74, 75),
                Total(Units.KWH),
                div10,
            ),
            "Total Battery Charge Energy": (
                pack_u16(76, 77),
                Total(Units.KWH),
                div10,
            ),
            "Today's Battery Discharge Energy": (78, DailyTotal(Units.KWH), div10),
            "Today's Battery Charge Energy": (79, DailyTotal(Units.KWH), div10),
            "Total PV Energy": (pack_u16(80, 81), Total(Units.KWH), div10),
            "Today's Energy": (82, DailyTotal(Units.KWH), div10),
            # 83-85: always 0
            "Total Feed-in Energy": (pack_u16(86, 87), Total(Units.KWH), div100),
            "Total Consumption": (pack_u16(88, 89), Total(Units.KWH), div100),
            "Today's Feed-in Energy": (90, DailyTotal(Units.KWH), div100),
            # 91: always 0
            "Today's Consumption": (92, DailyTotal(Units.KWH), div100),
            # 93-101: always 0
            # 102: always 1
            "Battery Remaining Capacity": (103, Units.PERCENT),
            # 104: always 1
            "Battery Temperature": (105, Units.C),
            "Battery Remaining Energy": (
                106,
                Measurement(Units.KWH, storage=True),
                div10,
            ),
            # 107: always 256 or 0
            # 108: always 3504
            # 109: always 2400
            # 110: around rise to 300 if battery not full, 0 if battery is full
            # 112, 113: range [250,350]; looks like 113 + offset = 112,
            #   peaks if battery is full
            # 114, 115: something around 33; Some temperature?!
            # 116: increases slowly [2,5]
            # 117-121: 1620	773	12850	12850	12850
            # 122-124: always 0
            # 125,126: some curve, look very similar to "42"(Battery Power),
            # with offset around 15
            # 127,128 resetting counter /1000, around battery charge + discharge
            # 164,165,166 some curves
            "Battery Operation mode": (168, Units.NONE, cls.Processors.battery_modes),
            # 169: div100 same as [39]
            # 170-199: always 0
        }

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][2]

    @classmethod
    def build_all_variants(cls, host, port, pwd=""):
        versions = [cls._build(host, port, pwd, False)]
        return versions



from .qvolt_hyb_g3_3p import QVOLTHYBG33P
from .x1 import X1
from .x1_boost import X1Boost
from .x1_hybrid_gen4 import X1HybridGen4
from .x1_mini import X1Mini
from .x1_mini_v34 import X1MiniV34
from .x1_smart import X1Smart
from .x3 import X3
from .x3_hybrid_g4 import X3HybridG4
from .x3_mic_pro_g2 import X3MicProG2
from .x3_ultra import X3Ultra
from .x3_v34 import X3V34
from .x_hybrid import XHybrid

__all__ = [
    "QVOLTHYBG33P",
    "XHybrid",
    "X1",
    "X1Mini",
    "X1MiniV34",
    "X1Smart",
    "X3V34",
    "X3HybridG4",
    "X3",
    "X1Boost",
    "X1HybridGen4",
    "X3MicProG2",
    "X3Ultra",
]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="solax",
    use_scm_version=True,
    author="Robin Wohlers-Reichel",
    author_email="me@robinwr.com",
    description="Solax inverter API client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/squishykid/solax",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "aiohttp>=3.5.4, <4",
        "voluptuous>=0.11.5",
        "importlib_metadata>=3.6; python_version<'3.10'",
        "typing_extensions>=4.1.0; python_version<'3.11'",
    ],
    setup_requires=[
        "setuptools_scm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    entry_points={
        "solax.inverter": [
            "qvolt_hyb_g3_3p = solax.inverters.qvolt_hyb_g3_3p:QVOLTHYBG33P",
            "x1 = solax.inverters.x1:X1",
            "x1_boost = solax.inverters.x1_boost:X1Boost",
            "x1_hybrid_gen4 = solax.inverters.x1_hybrid_gen4:X1HybridGen4",
            "x1_mini = solax.inverters.x1_mini:X1Mini",
            "x1_mini_v34 = solax.inverters.x1_mini_v34:X1MiniV34",
            "x1_smart = solax.inverters.x1_smart:X1Smart",
            "x3 = solax.inverters.x3:X3",
            "x3_hybrid_g4 = solax.inverters.x3_hybrid_g4:X3HybridG4",
            "x3_ultra = solax.inverters.x3_ultra:X3Ultra",
            "x3_mic_pro_g2 = solax.inverters.x3_mic_pro_g2:X3MicProG2",
            "x3_v34 = solax.inverters.x3_v34:X3V34",
            "x_hybrid = solax.inverters.x_hybrid:XHybrid",
        ],
    },
)



from typing import Any, Dict, Optional

import voluptuous as vol

from solax.inverter import Inverter
from solax.units import DailyTotal, Total, Units
from solax.utils import startswith


class X1(Inverter):
    # pylint: disable=duplicate-code
    _schema = vol.Schema(
        {
            vol.Required("type"): vol.All(str, startswith("X1-")),
            vol.Required("sn"): str,
            vol.Required("ver"): str,
            vol.Required("data"): vol.Schema(
                vol.All(
                    [vol.Coerce(float)],
                    vol.Any(
                        vol.Length(min=102, max=102),
                        vol.Length(min=103, max=103),
                        vol.Length(min=107, max=107),
                    ),
                )
            ),
            vol.Required("information"): vol.Schema(vol.All(vol.Length(min=9, max=9))),
        },
        extra=vol.REMOVE_EXTRA,
    )

    @classmethod
    def response_decoder(cls):
        return {
            "PV1 Current": (0, Units.A),
            "PV2 Current": (1, Units.A),
            "PV1 Voltage": (2, Units.V),
            "PV2 Voltage": (3, Units.V),
            "Output Current": (4, Units.A),
            "Network Voltage": (5, Units.V),
            "AC Power": (6, Units.W),
            "Inverter Temperature": (7, Units.C),
            "Today's Energy": (8, DailyTotal(Units.KWH)),
            "Total Energy": (9, Total(Units.KWH)),
            "Exported Power": (10, Units.W),
            "PV1 Power": (11, Units.W),
            "PV2 Power": (12, Units.W),
            "Battery Voltage": (13, Units.V),
            "Battery Current": (14, Units.A),
            "Battery Power": (15, Units.W),
            "Battery Temperature": (16, Units.C),
            "Battery Remaining Capacity": (21, Units.PERCENT),
            "Total Feed-in Energy": (41, Total(Units.KWH)),
            "Total Consumption": (42, Total(Units.KWH)),
            "Power Now": (43, Units.W),
            "Grid Frequency": (50, Units.HZ),
            "EPS Voltage": (53, Units.V),
            "EPS Current": (54, Units.A),
            "EPS Power": (55, Units.W),
            "EPS Frequency": (56, Units.HZ),
        }

    # pylint: enable=duplicate-code

    @classmethod
    def inverter_serial_number_getter(cls, response: Dict[str, Any]) -> Optional[str]:
        return response["information"][3]



import pytest
from voluptuous import Invalid

from solax.utils import contains_none_zero_value, startswith


def test_does_start_with():
    expected = "AAAA"
    actual = "AAAA"
    bingo = startswith(expected)(actual)
    assert bingo == actual


def test_does_not_start_with():
    expected = "AAAA"
    actual = "BBBB"
    with pytest.raises(Invalid):
        startswith(expected)(actual)


def test_is_not_str():
    expected = "AAAA"
    actual = 1
    with pytest.raises(Invalid):
        startswith(expected)(actual)


def test_contains_none_zero_value():
    with pytest.raises(Invalid):
        contains_none_zero_value([0])

    with pytest.raises(Invalid):
        contains_none_zero_value([0, 0])

    not_a_list = 1
    with pytest.raises(Invalid):
        contains_none_zero_value(not_a_list)

    expected = [1, 0]
    actual = contains_none_zero_value(expected)
    assert actual == expected

    expected = [-1, 1]
    actual = contains_none_zero_value(expected)
    assert actual == expected
